@@ -472,7 +472,8 @@ public abstract class AbstractEndpoint<S,U> {
     protected void releaseSSLContext(SSLHostConfig sslHostConfig) {
         for (SSLHostConfigCertificate certificate : sslHostConfig.getCertificates(true)) {
             if (certificate.getSslContext() != null) {
-                SSLContext sslContext = certificate.getSslContext();
+                // Only release the SSLContext if we generated it.
+                SSLContext sslContext = certificate.getSslContextGenerated();
                 if (sslContext != null) {
                     sslContext.destroy();
                 }
@@ -1163,8 +1164,8 @@ public abstract class AbstractEndpoint<S,U> {
                 // That causes delays when running the unit tests. Explicitly enableing linger but with a timeout of
                 // zero seconds seems to fix the issue.
                 s.setSoLinger(true, 0);
-                if (getLog().isDebugEnabled()) {
-                    getLog().debug("About to unlock socket for:" + unlockAddress);
+                if (getLog().isTraceEnabled()) {
+                    getLog().trace("About to unlock socket for:" + unlockAddress);
                 }
                 s.connect(unlockAddress,utmo);
                 if (getDeferAccept()) {
@@ -1180,8 +1181,8 @@ public abstract class AbstractEndpoint<S,U> {
                             "User-Agent: Tomcat wakeup connection\r\n\r\n");
                     sw.flush();
                 }
-                if (getLog().isDebugEnabled()) {
-                    getLog().debug("Socket unlock completed for:" + unlockAddress);
+                if (getLog().isTraceEnabled()) {
+                    getLog().trace("Socket unlock completed for:" + unlockAddress);
                 }
             }
             // Wait for up to 1000ms acceptor threads to unlock. Particularly
@@ -1317,6 +1318,7 @@ public abstract class AbstractEndpoint<S,U> {
 
     public abstract void bind() throws Exception;
     public abstract void unbind() throws Exception;
+
     public abstract void startInternal() throws Exception;
     public abstract void stopInternal() throws Exception;
 
